@@ -41,10 +41,13 @@ typedef struct
     NvU32  os_major_version;
     NvU32  os_minor_version;
     NvU32  os_build_number;
-    char * os_build_version_str;
-    char * os_build_date_plus_str;
+    const char * os_build_version_str;
+    const char * os_build_date_plus_str;
 }os_version_info;
 
+/* Each OS defines its own version of this opaque type */
+struct os_work_queue;
+
 /*
  * ---------------------------------------------------------------------------
  *
@@ -117,8 +120,8 @@ void        NV_API_CALL  os_free_spinlock            (void *);
 NvU64       NV_API_CALL  os_acquire_spinlock         (void *);
 void        NV_API_CALL  os_release_spinlock         (void *, NvU64);
 NV_STATUS   NV_API_CALL  os_get_address_space_info   (NvU64 *, NvU64 *, NvU64 *, NvU64 *);
-NV_STATUS   NV_API_CALL  os_queue_work_item          (void *);
-NV_STATUS   NV_API_CALL  os_flush_work_queue         (void);
+NV_STATUS   NV_API_CALL  os_queue_work_item          (struct os_work_queue *, void *);
+NV_STATUS   NV_API_CALL  os_flush_work_queue         (struct os_work_queue *);
 void        NV_API_CALL  os_register_compatible_ioctl    (NvU32, NvU32);
 void        NV_API_CALL  os_unregister_compatible_ioctl  (NvU32, NvU32);
 NV_STATUS   NV_API_CALL  os_alloc_mutex              (void **);
@@ -129,6 +132,7 @@ void        NV_API_CALL  os_release_mutex            (void *);
 void*       NV_API_CALL  os_alloc_semaphore          (NvU32);
 void        NV_API_CALL  os_free_semaphore           (void *);
 NV_STATUS   NV_API_CALL  os_acquire_semaphore        (void *);
+NV_STATUS   NV_API_CALL  os_cond_acquire_semaphore   (void *);
 NV_STATUS   NV_API_CALL  os_release_semaphore        (void *);
 BOOL        NV_API_CALL  os_semaphore_may_sleep      (void);
 NV_STATUS   NV_API_CALL  os_get_version_info         (os_version_info*);
@@ -164,6 +168,10 @@ NV_STATUS   NV_API_CALL  os_ipmi_connect             (NvU32 devIndex, NvU8 myAdd
 void        NV_API_CALL  os_ipmi_disconnect          (void *pOsPriv);
 NV_STATUS   NV_API_CALL  os_ipmi_send_receive_cmd    (void *pOsPriv, nvipmi_req_resp_t *pReq);
 NV_STATUS   NV_API_CALL  os_call_nv_vmbus            (NvU32, void *);
+NV_STATUS   NV_API_CALL  os_open_temporary_file      (void **);
+void        NV_API_CALL  os_close_file               (void *);
+NV_STATUS   NV_API_CALL  os_write_file               (void *, NvU8 *, NvU64, NvU64);
+NV_STATUS   NV_API_CALL  os_read_file                (void *, NvU8 *, NvU64, NvU64);
 
 /*
  * ---------------------------------------------------------------------------
@@ -183,6 +191,12 @@ NV_STATUS   NV_API_CALL  os_call_nv_vmbus            (NvU32, void *);
 void NV_API_CALL  out_string(const char *str);
 int  NV_API_CALL  nv_printf(NvU32 debuglevel, const char *printf_format, ...);
 
+#define NV_DEV_PRINTF(debuglevel, nv, format, ... ) \
+        nv_printf(debuglevel, "NVRM: GPU " NV_PCI_DEV_FMT ": " format, NV_PCI_DEV_FMT_ARGS(nv), ## __VA_ARGS__)
+
+#define NV_DEV_PRINTF_STATUS(debuglevel, nv, status, format, ... ) \
+        nv_printf(debuglevel, "NVRM: GPU " NV_PCI_DEV_FMT ": " format " (0x%x)\n", NV_PCI_DEV_FMT_ARGS(nv), ## __VA_ARGS__, status)
+
 /*
  * those NV_MEMORY_* and NV_PROTECT_* have dup defines in osCore.h
  */
