@@ -16,12 +16,16 @@
 
 #include "nv-time.h"
 
+extern char *NVreg_TemporaryFilePath;
+
 #define MAX_ERROR_STRING 512
 static char nv_error_string[MAX_ERROR_STRING];
 nv_spinlock_t nv_error_string_lock;
 
 extern nv_linux_state_t nv_ctl_device;
 
+extern nv_kthread_q_t nv_kthread_q;
+
 NV_STATUS NV_API_CALL os_disable_console_access(void)
 {
     NV_ACQUIRE_CONSOLE_SEM();
@@ -177,6 +181,25 @@ NV_STATUS NV_API_CALL os_acquire_semaphore
     return NV_OK;
 }
 
+NV_STATUS NV_API_CALL os_cond_acquire_semaphore
+(
+    void * pSema
+)
+{
+    os_semaphore_t *os_sema = (os_semaphore_t *)pSema;
+    if (!NV_MAY_SLEEP())
+    {
+        return NV_ERR_INVALID_REQUEST;
+    }
+
+    if (down_trylock(os_sema))
+    {
+        return NV_ERR_TIMEOUT_RETRY;
+    }
+
+    return NV_OK;
+}
+
 NV_STATUS NV_API_CALL os_release_semaphore
 (
     void *pSema
@@ -970,51 +993,68 @@ NV_STATUS NV_API_CALL os_schedule(void)
     }
 }
 
-static void os_execute_work_item(
-    NV_WORKQUEUE_DATA_T *data
-)
+typedef struct {
+    nv_kthread_q_item_t item;
+    void *data;
+} os_queue_data_t;
+
+static void os_execute_work_item(void *_oqd)
 {
-    nv_work_t *work = NV_WORKQUEUE_UNPACK_DATA(data);
+    os_queue_data_t *oqd = _oqd;
     nvidia_stack_t *sp = NULL;
+    void *data = oqd->data;
+
+    NV_KFREE(oqd, sizeof(os_queue_data_t));
 
     if (nv_kmem_cache_alloc_stack(&sp) != 0)
     {
         return;
     }
 
-    rm_execute_work_item(sp, work->data);
-
-    os_free_mem((void *)work);
+    rm_execute_work_item(sp, data);
 
     nv_kmem_cache_free_stack(sp);
 }
 
-NV_STATUS NV_API_CALL os_queue_work_item(
-    void *nv_work
-)
+NV_STATUS NV_API_CALL os_queue_work_item(struct os_work_queue *queue, void *data)
 {
-    NV_STATUS status;
-    nv_work_t *work;
+    os_queue_data_t *oqd;
+    nv_kthread_q_t *kthread;
 
-    status = os_alloc_mem((void **)&work, sizeof(nv_work_t));
+    /* Use the global queue unless a valid queue was provided */
+    kthread = queue ? &queue->nvk : &nv_kthread_q;
 
-    if (NV_OK != status)
-        return status;
+    /* Make sure the kthread is active */
+    if (unlikely(!kthread->q_kthread)) {
+        nv_printf(NV_DBG_ERRORS, "NVRM: queue is not enabled\n");
+        return NV_ERR_NOT_READY;
+    }
 
-    work->data = nv_work;
+    /* Allocate atomically just in case we're called in atomic context. */
+    NV_KMALLOC_ATOMIC(oqd, sizeof(os_queue_data_t));
+    if (!oqd)
+        return NV_ERR_NO_MEMORY;
 
-    NV_WORKQUEUE_INIT(&work->task, os_execute_work_item,
-                      (void *)work);
-    NV_WORKQUEUE_SCHEDULE(&work->task);
+    nv_kthread_q_item_init(&oqd->item, os_execute_work_item, oqd);
+    oqd->data = data;
+
+    nv_kthread_q_schedule_q_item(kthread, &oqd->item);
 
     return NV_OK;
 }
 
-NV_STATUS NV_API_CALL os_flush_work_queue(void)
+NV_STATUS NV_API_CALL os_flush_work_queue(struct os_work_queue *queue)
 {
+    nv_kthread_q_t *kthread;
+
+    /* Use the global queue unless a valid queue was provided */
+    kthread = queue ? &queue->nvk : &nv_kthread_q;
+
     if (NV_MAY_SLEEP())
     {
-        NV_WORKQUEUE_FLUSH();
+        if (kthread->q_kthread)
+            nv_kthread_q_flush(kthread);
+
         return NV_OK;
     }
     else
@@ -1226,32 +1266,11 @@ NV_STATUS NV_API_CALL os_get_version_info(os_version_info * pOsVersionInfo)
     pOsVersionInfo->os_build_number  = NV_KERNEL_SUBVERSION;
 
 #if defined(UTS_RELEASE)
-    do
-    {
-        char * version_string = NULL;
-        status = os_alloc_mem((void **)&version_string,
-                              (strlen(UTS_RELEASE) + 1));
-        if (status != NV_OK)
-        {
-            return status;
-        }
-        strcpy(version_string, UTS_RELEASE);
-        pOsVersionInfo->os_build_version_str = version_string;
-    }while(0);
+    pOsVersionInfo->os_build_version_str = UTS_RELEASE;
 #endif
 
 #if defined(UTS_VERSION)
-    do
-    {
-        char * date_string    = NULL;
-        status = os_alloc_mem((void **)&date_string, (strlen(UTS_VERSION) + 1));
-        if (status != NV_OK)
-        {
-            return status;
-        }
-        strcpy(date_string, UTS_VERSION);
-        pOsVersionInfo->os_build_date_plus_str = date_string;
-    }while(0);
+    pOsVersionInfo->os_build_date_plus_str = UTS_VERSION;
 #endif
 
     return status;
@@ -1603,6 +1622,18 @@ NV_STATUS NV_API_CALL os_alloc_pages_node
             nid, order);
         status = NV_ERR_NO_MEMORY;
     }
+    else if (page_to_nid(alloc_addr) != nid)
+    {
+        // 
+        // We can hit this case when a Linux kernel bug is not patched.
+        // The needed patch is https://patchwork.kernel.org/patch/10427387/
+        //
+        nv_printf(NV_DBG_ERRORS,
+            "NVRM: alloc_pages_node(node = %d, order = %u) wrong node ID.\n",
+            nid, order);
+        __free_pages(alloc_addr, order);
+        status = NV_ERR_NO_MEMORY;
+    }
     else
     {
         *pAddress = (NvU64)page_to_phys(alloc_addr);
@@ -1644,16 +1675,9 @@ NvU32 NV_API_CALL os_count_tail_pages
     NvU64 address
 )
 {
-#if defined(NV_COMPOUND_ORDER_PRESENT)
-    // 
-    // Since both compound_head and compound_order are checked in under the 
-    // same commit, only one contest is used.
-    //
     NvU32 order = compound_order(compound_head(NV_GET_PAGE_STRUCT(address)));
-    return (1 << order);
-#else
-    return 1;
-#endif
+
+    return 1 << order;
 }
 
 void NV_API_CALL os_free_pages_phys
@@ -1678,28 +1702,6 @@ struct nv_ipmi_softc
     NvU32               seqNum;     //request sequence number
 };
 
-static inline int
-nv_ipmi_set_my_address
-(
-    struct ipmi_user    *user,
-    unsigned char       address
-)
-{
-#if defined(IPMICTL_SET_MY_CHANNEL_ADDRESS_CMD)
-    //
-    // Commit c14979b993021377228958498937bcdd9539cbce
-    // has introduced new IPMI ioctl commands, that allow to specify
-    // per-channel IPMB addresses. Consequently ipmi_set_my_address()
-    // has received the additional channel argument, and started
-    // returning a status (-EINVAL for invalid channel).
-    //
-    return ipmi_set_my_address(user, 0, address);
-#else
-    ipmi_set_my_address(user, address);
-    return 0;
-#endif
-}
-
 static void
 nv_ipmi_receive_handler
 (
@@ -1774,7 +1776,7 @@ NV_STATUS NV_API_CALL os_ipmi_connect
         goto quit;
     }
 
-    err_no = nv_ipmi_set_my_address(p_priv->p_user, myAddr);
+    err_no = ipmi_set_my_address(p_priv->p_user, 0, myAddr);
 
     if (err_no < 0)
     {
@@ -2037,3 +2039,129 @@ NV_STATUS NV_API_CALL os_call_nv_vmbus(NvU32 vmbus_cmd, void *input)
     return NV_ERR_NOT_SUPPORTED;
 }
 #endif
+
+NV_STATUS NV_API_CALL os_open_temporary_file
+(
+    void **ppFile
+)
+{
+#if defined(O_TMPFILE) && defined(NV_KERNEL_WRITE_PRESENT)
+    struct file *file;
+    const char *path = NVreg_TemporaryFilePath;
+
+    if (!path)
+        path = "/tmp";
+
+    file = filp_open(path, O_TMPFILE | O_LARGEFILE | O_RDWR, 0);
+    if (IS_ERR(file))
+    {
+        return NV_ERR_OPERATING_SYSTEM;
+    }
+
+    *ppFile = (void *)file;
+
+    return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
+}
+
+void NV_API_CALL os_close_file
+(
+    void *pFile
+)
+{
+#if defined(O_TMPFILE) && defined(NV_KERNEL_WRITE_PRESENT)
+    filp_close(pFile, NULL);
+#endif
+}
+
+#define NV_MAX_NUM_FILE_IO_RETRIES 10
+
+NV_STATUS NV_API_CALL os_write_file
+(
+    void *pFile,
+    NvU8 *pBuffer,
+    NvU64 size,
+    NvU64 offset
+)
+{
+#if defined(O_TMPFILE) && defined(NV_KERNEL_WRITE_PRESENT)
+    loff_t f_pos = offset;
+    ssize_t num_written;
+    int num_retries = NV_MAX_NUM_FILE_IO_RETRIES;
+
+retry:
+#if defined(NV_KERNEL_WRITE_HAS_POINTER_POS_ARG)
+    num_written = kernel_write(pFile, pBuffer, size, &f_pos);
+#else
+    num_written = kernel_write(pFile, pBuffer, size, f_pos);
+#endif
+    if (num_written < 0)
+    {
+        return NV_ERR_OPERATING_SYSTEM;
+    }
+    else if (num_written < size)
+    {
+        if (num_written > 0)
+        {
+            pBuffer += num_written;
+            size -= num_written;
+        }
+        if (--num_retries > 0)
+        {
+            cond_resched();
+            goto retry;
+        }
+        return NV_ERR_OPERATING_SYSTEM;
+    }
+
+    return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
+}
+
+NV_STATUS NV_API_CALL os_read_file
+(
+    void *pFile,
+    NvU8 *pBuffer,
+    NvU64 size,
+    NvU64 offset
+)
+{
+#if defined(O_TMPFILE) && defined(NV_KERNEL_READ_PRESENT)
+    loff_t f_pos = offset;
+    ssize_t num_read;
+    int num_retries = NV_MAX_NUM_FILE_IO_RETRIES;
+
+retry:
+#if defined(NV_KERNEL_READ_HAS_POINTER_POS_ARG)
+    num_read = kernel_read(pFile, pBuffer, size, &f_pos);
+#else
+    num_read = kernel_read(pFile, f_pos, pBuffer, size);
+#endif
+    if (num_read < 0)
+    {
+        return NV_ERR_OPERATING_SYSTEM;
+    }
+    else if (num_read < size)
+    {
+        if (num_read > 0)
+        {
+            pBuffer += num_read;
+            size -= num_read;
+        }
+        if (--num_retries > 0)
+        {
+            cond_resched();
+            goto retry;
+        }
+        return NV_ERR_OPERATING_SYSTEM;
+    }
+
+    return NV_OK;
+#else
+    return NV_ERR_NOT_SUPPORTED;
+#endif
+}
